/*
 *  Copyright (c) xlightweb.org, 2008. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xlightweb.org/
 */
package org.xlightweb;

import java.io.IOException;
import java.nio.BufferUnderflowException;
import java.nio.ByteBuffer;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.xlightweb.ComposedByteBuffer.IBufferSink;
import org.xsocket.MaxReadSizeExceededException;



/**
 * body parser implementation base
 * 
 * @author grro@xlightweb.org
 */
abstract class AbstractBodyParser implements IBufferSink {

	
	private static final Logger LOG = Logger.getLogger(AbstractBodyParser.class.getName());
	
	private final AbstractHttpConnection httpConnection;
	
	private final IHttpHeader header;
	private final NonBlockingBodyDataSource dataSource;
	
	private boolean isComplete = false;
	

	/**
	 * constructor 
	 * 
	 * @param httpConnection     the  http connection
	 * @param header             the header 
	 * @throws IOException if an exception occurs
	 */
	public AbstractBodyParser(AbstractHttpConnection httpConnection, IHttpHeader header) throws IOException {
		this.httpConnection = httpConnection;
		this.header = header;
		
		this.dataSource = new NonBlockingBodyDataSource(header.getCharacterEncoding(), httpConnection, httpConnection.getExecutor());
	}



	/**
	 * gets the header 
	 * 
	 * @return the header
	 */
	final IHttpHeader getHeader() {
		return header;
	}
	
	
	/**
	 * gets the connection id
	 * @return the connection id
	 */
	final String getConnectionId() {
		return httpConnection.getId();
	}
	
	
	/**
	 * appends data data to the body  
	 */
	public final void add(ByteBuffer buffer) throws IOException {
		dataSource.append(buffer);
	}
	
	
	/**
	 * register an exception 
	 * 
	 * @param ioe  the exception
	 */
	final void register(IOException ioe)  {
		dataSource.setIOException(ioe);
	}
	
	
	/**
	 * returns the body data source
	 * 
	 * @return the body data source
	 */
	NonBlockingBodyDataSource getDataSource() {
		return dataSource;
	}
	
	
	
	

	/**
	 * parses the the body   
	 *  
	 * @param rawData   the raw read network data buffer 
	 * @return true, if the body is complete
	 * @throws IOException  if an exception occurs 
	 * @throws BufferUnderflowException if more data is required
	 * @throws MaxReadSizeExceededException if the max read size is exceeded
	 */
	final boolean parse(ComposedByteBuffer rawData) throws IOException, BufferUnderflowException, MaxReadSizeExceededException {			
		
		if (!dataSource.isComplete()) {
				
			try {
				doParse(rawData);
					
				if (isComplete) {
					return true;
				} else {
					return false;
				}
					
			} catch (BufferUnderflowException bue) {
				throw bue;
					
			} catch (IOException ioe) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("[" + httpConnection.getId() + "] (protocol?) error occured by reading body " + ioe.toString());
				}
	
				dataSource.setIOException(ioe);
				httpConnection.destroy();
				dataSource.close();
					
				throw ioe;
			}
		} 
		
		return false;
	}
	
	

	
	/**
	 * parse the body 
	 * @param rawData       the raw read network data buffer
	 * @throws IOException if an exception occurs  
	 */
	abstract void doParse(ComposedByteBuffer rawData) throws IOException;


	
	/**
	 * error notification 
	 * 
	 * 
	 * @param ioe      the error
	 * @param rawData  the unprocessed raw data 
	 */
	abstract void onException(IOException ioe, ComposedByteBuffer rawData);
	
	
	/**
	 * set the body as received
	 */
	final void setComplete() {
		isComplete = true;
		dataSource.setComplete(true);
	}

	/**
	 * return true if is complete
	 * @return true if is complete
	 */
	final boolean isComplete() {
		return isComplete;
	}
	

	/**
	 * disconnect handling
	 */
	void onDisconnect() {
		dataSource.onDisconnect();
	}
}
