/*
 *  Copyright (c) xlightweb.org, 2008. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xlightweb.org/
 */
package org.xlightweb;


import java.io.IOException;
import java.nio.channels.ClosedChannelException;
import java.util.logging.Level;
import java.util.logging.Logger;



/**
 * a full body parser
 *  
 * @author grro@xlightweb.org
 */
final class FullMessageBodyParser extends AbstractBodyParser {

	private static final Logger LOG = Logger.getLogger(FullMessageBodyParser.class.getName());
	
	private final int size;
	private int remaining;
	
	
	/**
	 * constructor 
	 * 
	 * @param httpConnection   the http connection
	 * @param header           the header
	 * @throws IOException if an exception occurs
	 */
	public FullMessageBodyParser(AbstractHttpConnection httpConnection, IHttpHeader header) throws IOException {
		super(httpConnection, header);
		
		size = header.getContentLength();
		remaining = size;
	}

	

	/**
	 * {@inheritDoc}
	 */
	@Override
	void doParse(ComposedByteBuffer rawData) throws IOException {

		int available = rawData.available();
		
		if (available == 0) {
			return;
		}
		
		
		// not enough data
		if (available < remaining) {
			rawData.readByteBufferByLength(available, this);
			remaining -= available;

			if (LOG.isLoggable(Level.FINE)) {
				LOG.fine("[" + getConnectionId() + "] " + available + " bytes read remaining=" + remaining);
			}

		// enough data 
		} else {
			if (LOG.isLoggable(Level.FINE)) {
				LOG.fine("[" + getConnectionId() + "] reading remaining data " + remaining);
			}
			
			
			rawData.readByteBufferByLength(remaining, this);
			remaining = 0;
		} 

		if (remaining == 0) {
			if (LOG.isLoggable(Level.FINE)) {
				LOG.fine("[" + getConnectionId() + "] complete body received (size=" + size + ")");
			}
			
			setComplete();
			
		} else {
			if (LOG.isLoggable(Level.FINER)) {
				LOG.finer("[" + getConnectionId() + "] body data read. wating for more data");
			}

		}
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	@Override
	void onException(IOException ioe, ComposedByteBuffer rawData) {
		if (!isComplete()) {
			String msg = "error occured by parsing full message body (expected size: " + size + " received size: " + (size - remaining) + ") ";
			if (ioe instanceof ClosedChannelException) {
				msg = msg + "connection disconnected (by peer?)";
			} else {
				msg = msg + ioe.toString();
			}
	        register(new ProtocolException(msg));
		}
	}
}

 