/*
 *  Copyright (c) xlightweb.org, 2008. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xlightweb.org/
 */
package org.xlightweb;


import java.io.IOException;
import java.nio.channels.ClosedChannelException;

import org.xlightweb.HeaderlineParser.IHeaderSink;




/**
 * chunked body parser 
 *  
 * @author grro@xlightweb.org
 */
final class FullMessageChunkedBodyParser extends AbstractBodyParser implements IHeaderSink {
    

	private static final byte CR = 13;
	private static final byte LF = 10;
	private static final byte SPACE = 32;
	private static final byte HTAB = 9;

	
	private static final int STATE_READ_LENGTH_FIELD = 0;
	private static final int STATE_READ_CONTENT = 1;
	private static final int STATE_READ_CONTENT_CRLF = 2;
	private static final int STATE_READ_TRAILER = 3;
	private static final int STATE_COMPLETE = 999;
	
	private int state = 0;
	
	private final StringBuilder stringBuilder = new StringBuilder(8); 
	
	private int remainingDataToRead;
	
	private final IHttpHeader header;
	private HeaderlineParser headerlineParser;
		
	
	
	/**
	 * constructor
	 * 
	 * @param httpConnection   the http connection
	 * @param header           the header
	 * @throws IOException if an exception occurs
	 */
	public FullMessageChunkedBodyParser(AbstractHttpConnection httpConnection, IHttpHeader header) throws IOException {
		super(httpConnection, header);
		
		this.header = header;
	}

	
	

	/**
	 * {@inheritDoc}
	 */
	@Override
	void doParse(ComposedByteBuffer rawData) throws IOException {
		

		while (rawData.available() > 0) {
			
			switch (state) {
			
				case STATE_READ_LENGTH_FIELD:
					byte b = rawData.getByte();
					switch (b) {
						case CR:
							break;
							
						case SPACE:
							break;
							
						case HTAB:
							break;
							
						case LF:
							remainingDataToRead = parseLengtField(stringBuilder.toString());
							stringBuilder.setLength(0);
							
							if (remainingDataToRead > 0) {
								state = STATE_READ_CONTENT;
								
							} else {
								state = STATE_READ_TRAILER;
								headerlineParser = HeaderlineParser.newInstance();
							}
							break;
							
						default:
							stringBuilder.append((char) b);
							break;
						}
					break;
		
					
				case STATE_READ_CONTENT:
					
					int size = rawData.remaining();
					if (size >= remainingDataToRead) {
						size = remainingDataToRead;
						remainingDataToRead = 0;
						
					} else {
						remainingDataToRead -= size;
					}
					rawData.readByteBufferByLength(size, this);
					

					if (remainingDataToRead == 0) {
						state = STATE_READ_CONTENT_CRLF;
					} 
					break;
					
					
				case STATE_READ_CONTENT_CRLF:
					
					b = rawData.getByte();
					switch (b) {
						case CR:
							break;
						
						case LF:
							state = STATE_READ_LENGTH_FIELD;
							break;
		
						default:
							break;
						}
					break;
					
					
				case STATE_READ_TRAILER:
					boolean isRead = headerlineParser.parse(rawData, this);
					if (isRead) {
						state = STATE_COMPLETE;
						setComplete();
						HeaderlineParser.recycleInstance(headerlineParser);
						return;
					}
					break;
					
					
				default:
					break;
			}
		}
	}
	
	
	   
    /**
     * {@inheritDoc}
     */
    @Override
    void onException(IOException ioe, ComposedByteBuffer rawData) {
    	if (!isComplete()) {
	    	String msg = "error occured by parsing full message chunked body (remaing size of current chunk: " + remainingDataToRead + ") ";
	    	if (ioe instanceof ClosedChannelException) {
				msg = msg + "connection disconnected (by peer?)";
			} else {
				msg = msg + ioe.toString();
			}
	    	register(new ProtocolException(msg));
    	}
    }
    

	/**
	 * add a header
	 * 
	 * @param headername    the header name
	 * @param headervalue   the header value
	 */
	public void addHeader(String headername, String headervalue) {
		header.addHeader(headername, headervalue);
	}
	
	
	private int parseLengtField(String lengthField) throws IOException {
		try {
			return Integer.parseInt(lengthField, 16);
			
		} catch (NumberFormatException nfe) {

			// chunk extension?
			if (lengthField.indexOf(";") != -1) {
				String length	= lengthField.substring(0, lengthField.indexOf(";"));
				return Integer.parseInt(length, 16);
				
			} else {
				throw new IOException("[" + getConnectionId() + "] http protocol error. length field expected. Got " + lengthField);
			}
		}
	}	
}

 