/*
 *  Copyright (c) xlightweb.org, 2008. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xlightweb.org/
 */
package org.xlightweb;

import java.io.IOException;
import java.net.ConnectException;
import java.nio.ByteBuffer;

import org.xlightweb.AbstractHttpConnection.IMultimodeExecutor;
import org.xsocket.Execution;
import org.xsocket.connection.IWriteCompletionHandler;
import org.xsocket.connection.IConnection.FlushMode;




 

/**
 * Provides an abstract class to be subclassed to implement a 
 * {@link IHttpRequestHandler}. A subclass of <code>HttpRequestHandlerAdapter</code> 
 * must override at least one method, usually one of these:
 *
 * <ul>
 *    <li> <code>doGet</code>, if the implementation supports HTTP GET requests
 *    <li> <code>doPost</code>, for HTTP POST requests
 *    <li> <code>doPut</code>, for HTTP PUT requests
 *    <li> <code>doDelete</code>, for HTTP DELETE requests 
 * </ul>
 * 
 * If the method for a dedicated HTTP method is not sub classes, 
 * a 405 error will be returned. <br> <br>
 * 
 * Example:
 * <pre>
 * class MyRequestHandler extends HttpRequestHandler {
 * 
 *    &#064Override
 *    protected void doGet(IHttpExchange exchange) throws IOException, BadMessageException {
 *       exchange.send(new HttpResponse(200, "text/plain", "GET called"));
 *    }
 *    
 *    &#064Override
 *    protected void doPost(IHttpExchange exchange) throws IOException, BadMessageException {
 *       exchange.send(new HttpResponse(200, "text/plain", "POST called"));
 *    }
 * }
 * </pre>
 * 
 * For more examples see {@link IHttpRequestHandler}
 * 
 *
 * @author grro@xlightweb.org
 */
public class HttpRequestHandler implements IHttpRequestHandler {

	
	private static final NullBodyWriter NULL_BODYWRITER = new NullBodyWriter();
	private static final MultimodeExecutor MULTIMODE_EXECUTOR = new MultimodeExecutor(); 
	
	
	/**
	 * {@inheritDoc}
	 */
	public final void onRequest(IHttpExchange exchange) throws IOException {
		
		String method = exchange.getRequest().getMethod();
		
		if (method.equals(IHttpRequest.GET_METHOD)) {
			doGet(exchange);
			
		} else if (method.equals(IHttpRequest.POST_METHOD)) {
			doPost(exchange);

		} else if (method.equals(IHttpRequest.PUT_METHOD)) {
			doPut(exchange);

		} else if (method.equals(IHttpRequest.DELETE_METHOD)) {
			doDelete(exchange);

		} else if (method.equals(IHttpRequest.HEAD_METHOD)) {
			doHead(exchange);
			
		} else if (method.equals(IHttpRequest.OPTIONS_METHOD)) {
			doOptions(exchange);

		} else if (method.equals(IHttpRequest.TRACE_METHOD)) {
			doTrace(exchange);

		} else {
			sendMethodNotAllowed(exchange, method);
		}
	}
	
	
	
	/**
     * call back which will be called, if a GET request is received 
     * 
     * <p>Overriding this method to support a GET request also
     * automatically supports an HTTP HEAD request. A HEAD
     * request is a GET request that returns no body in the
     * response, only the request header fields. </p>
     * 
     * <p>The GET method should also be idempotent, meaning
     * that it can be safely repeated. </p>
     * 
	 * @param exchange  the exchange
	 * @throws IOException if an exception occurred. By throwing this exception an error http response message 
	 *                     will be sent by xSocket, if one or more requests are unanswered. The underlying 
	 *                     connection will be closed
	 * @throws BadMessageException By throwing this exception an error http response message will be sent by xSocket,
	 *                             which contains the exception message. The underlying connection will be closed                 
	 */
	protected void doGet(IHttpExchange exchange) throws IOException, BadMessageException {
		sendMethodNotAllowed(exchange, "");
	}

	
	/**
     * call back which will be called, if a POST request is received
     * 
     * <p>This method does not need to be either safe or idempotent.
     * Operations requested through POST can have side effects</p>
     * 
	 * @param exchange  the exchange
	 * @throws IOException if an exception occurred. By throwing this exception an error http response message 
	 *                     will be sent by xSocket, if one or more requests are unanswered. The underlying 
	 *                     connection will be closed
	 * @throws BadMessageException By throwing this exception an error http response message will be sent by xSocket,
	 *                             which contains the exception message. The underlying connection will be closed                 
	 */
	protected void doPost(IHttpExchange exchange) throws IOException, BadMessageException {  
		sendMethodNotAllowed(exchange, "POST");
	}
	
	
	/**
     * call back which will be called, if a PUT request is received
     * 
     * <p>This method does not need to be either safe or idempotent.
     * Operations that <code>doPut</code> performs can have side
     * effects</p>
     * 
	 * @param exchange  the exchange
	 * @throws IOException if an exception occurred. By throwing this exception an error http response message 
	 *                     will be sent by xSocket, if one or more requests are unanswered. The underlying 
	 *                     connection will be closed
	 * @throws BadMessageException By throwing this exception an error http response message will be sent by xSocket,
	 *                             which contains the exception message. The underlying connection will be closed                 
	 */
	protected void doPut(IHttpExchange exchange) throws IOException, BadMessageException {  
		sendMethodNotAllowed(exchange, "PUT");
	}
	
	
	/**
     * call back which will be called, if a DELETE request is received
     * 
     * <p>This method does not need to be either safe
     * or idempotent. Operations requested through
     * DELETE can have side effects</p>
     * 
	 * @param exchange  the exchange
	 * @throws IOException if an exception occurred. By throwing this exception an error http response message 
	 *                     will be sent by xSocket, if one or more requests are unanswered. The underlying 
	 *                     connection will be closed
	 * @throws BadMessageException By throwing this exception an error http response message will be sent by xSocket,
	 *                             which contains the exception message. The underlying connection will be closed                 
	 */
	protected void doDelete(IHttpExchange exchange) throws IOException, BadMessageException {  
		sendMethodNotAllowed(exchange, "DELETE");
	}
	
	
	/**
     * call back which will be called, if a HEAD request is received
     * 
     * <p>The client sends a HEAD request when it wants
     * to see only the headers of a response, such as
     * Content-Type or Content-Length.</p>
     * 
	 * @param exchange  the exchange
	 * @throws IOException if an exception occurred. By throwing this exception an error http response message 
	 *                     will be sent by xSocket, if one or more requests are unanswered. The underlying 
	 *                     connection will be closed
	 * @throws BadMessageException By throwing this exception an error http response message will be sent by xSocket,
	 *                             which contains the exception message. The underlying connection will be closed                 
	 */
	protected void doHead(IHttpExchange exchange) throws IOException, BadMessageException {
		doGet(new HeadMethodHttpExchange(exchange));
	}

	
	
	/**
     * call back which will be called, if a OPTIONS request is received
     *
	 * @param exchange  the exchange
	 * @throws IOException if an exception occurred. By throwing this exception an error http response message 
	 *                     will be sent by xSocket, if one or more requests are unanswered. The underlying 
	 *                     connection will be closed
	 * @throws BadMessageException By throwing this exception an error http response message will be sent by xSocket,
	 *                             which contains the exception message. The underlying connection will be closed                 
	 */
	protected void doOptions(IHttpExchange exchange) throws IOException, BadMessageException {  
		sendMethodNotAllowed(exchange, "OPTIONS");
	}

	
	/**
     * call back which will be called, if a TRACE request is received
     * 
     *
	 * @param exchange  the exchange
	 * @throws IOException if an exception occurred. By throwing this exception an error http response message 
	 *                     will be sent by xSocket, if one or more requests are unanswered. The underlying 
	 *                     connection will be closed
	 * @throws BadMessageException By throwing this exception an error http response message will be sent by xSocket,
	 *                             which contains the exception message. The underlying connection will be closed                 
	 */
	protected void doTrace(IHttpExchange exchange) throws IOException, BadMessageException {  
		sendMethodNotAllowed(exchange, "TRACE");
	}


	
	private void sendMethodNotAllowed(IHttpExchange exchange, String method) {
		exchange.sendError(405, "Http method " + method + " is not supported by this URL");
	}
	
	
	private static final class HeadMethodHttpExchange implements IHttpExchange {

		private IHttpExchange delegee = null;
		
		public HeadMethodHttpExchange(IHttpExchange delegee) {
			this.delegee = delegee;
		}
		
		public void destroy() {
			delegee.destroy();
		}

		public BodyDataSink forward(IHttpRequestHeader requestHeader, IHttpResponseHandler responseHandler) throws IOException, ConnectException, IllegalStateException {
			return delegee.forward(requestHeader, responseHandler);
		}

		
		public BodyDataSink forward(IHttpRequestHeader requestHeader) throws IOException, ConnectException, IllegalStateException {
			return delegee.forward(requestHeader);
		}
		
		
		public BodyDataSink forward(IHttpRequestHeader requestHeader, int contentLength) throws IOException, ConnectException, IllegalStateException {
			return delegee.forward(requestHeader, contentLength);
		}
		
		
		public BodyDataSink forward(IHttpRequestHeader requestHeader, int contentLength, IHttpResponseHandler responseHandler) throws IOException, ConnectException, IllegalStateException {
			return delegee.forward(requestHeader, contentLength, responseHandler);
		}

		public void forward(IHttpRequest request) throws IOException, ConnectException, IllegalStateException {
			delegee.forward(request);
		}

		public void forward(IHttpRequest request, IHttpResponseHandler responseHandler) throws IOException, ConnectException, IllegalStateException {
			delegee.forward(request, responseHandler);
		}

		public IHttpConnection getConnection() {
			return delegee.getConnection();
		}

		public IHttpRequest getRequest() {
			return delegee.getRequest();
		}

		public IHttpSession getSession(boolean create) {
			return delegee.getSession(create);
		}
		
		public String encodeURL(String url) {
			return delegee.encodeURL(url);
		}

		
		public BodyDataSink send(final IHttpResponseHeader header) throws IOException, IllegalStateException {
		
			final BodyDataSink bodyDataSink = newEmtpyBodyDataSink();
			
			IBodyCloseListener closeListener = new IBodyCloseListener() {
				
				@Execution(Execution.NONTHREADED)
				public void onClose() throws IOException {
					delegee.send(new HttpResponse(header));
				}
			};
			
			bodyDataSink.addCloseListener(closeListener);
			return bodyDataSink;
		}

		
		public BodyDataSink send(IHttpResponseHeader header, int contentLength) throws IOException, IllegalStateException {
			return send(header);
		}

		
		public void send(IHttpResponse response) throws IOException, IllegalStateException {
			delegee.send(new HttpResponse(response.getResponseHeader()));
		}

		public void sendError(int errorCode, String msg) throws IllegalStateException {
			delegee.sendError(errorCode, msg);
		}

		public void sendError(int errorCode) throws IllegalStateException {
			delegee.sendError(errorCode);
		}

		public void sendError(Exception e) throws IllegalStateException {
			delegee.sendError(e);
		}
		
		
		private BodyDataSink newEmtpyBodyDataSink() throws IOException {

			return new BodyDataSink((AbstractHttpConnection) getConnection(), MULTIMODE_EXECUTOR, NULL_BODYWRITER, "UTF-8");
		}
	}	
	
	
	private static final class NullBodyWriter implements IBodyWriter {
		
	    public void flush(ByteBuffer[] bodyData, FlushMode flushMode, IWriteCompletionHandler completionHandler) throws IOException {
	        if (completionHandler != null) {
	            completionHandler.onWritten(0);
	        }
		}
		
		public void close() throws IOException {
		}
		
		public void destroy() {
		}
	}
	
	
	private static final class MultimodeExecutor implements IMultimodeExecutor {
		
		public void processMultithreaded(Runnable task) {
			task.run();
		}
		
		public void processNonthreaded(Runnable task) {
			task.run();
		}
	}
}
