/*
 *  Copyright (c) xlightweb.org, 2008. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xlightweb.org/
 */
package org.xlightweb.client;

import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TimerTask;
import java.util.UUID;
import java.util.Map.Entry;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.xlightweb.IHttpSession;



/**
 * Session Manager
 * 
 * @author grro@xlightweb.org
 */
final class SessionManager {
	
	public static final Logger LOG = Logger.getLogger(SessionManager.class.getName());

	private static final List<WeakReference<SessionManager>> INSTANCE_LIST = Collections.synchronizedList(new ArrayList<WeakReference<SessionManager>>());  
	
	
	private static final long CLEANING_PERIOD_MILLIS = 5 * 1000;
	static {
		TimerTask cleaningTask = new TimerTask() {
			@Override
			public void run() {
				for (SessionManager sessionManager : getInstances()) {
					sessionManager.clean();
				}
			}
		};
		HttpClientConnection.schedule(cleaningTask, CLEANING_PERIOD_MILLIS, CLEANING_PERIOD_MILLIS);
	}
	
	
	private final HashMap<DomainPathPair, HttpSession> sessions = new HashMap<DomainPathPair, HttpSession>();

	
	

	/**
	 * constructor
	 */
	public SessionManager() {
		WeakReference<SessionManager> sessionManagerRef = new WeakReference<SessionManager>(this);
		INSTANCE_LIST.add(sessionManagerRef);
	}
	
	
	

	private static List<SessionManager> getInstances() {
		List<SessionManager> instances = new ArrayList<SessionManager>();
		
		boolean isCleanRequired = false;
		
		for (WeakReference<SessionManager> sessionManagerRef : INSTANCE_LIST) {
			SessionManager sessionManager = sessionManagerRef.get();
			if (sessionManager != null) {
				instances.add(sessionManager);
			} else {
				isCleanRequired = true;
			}
		}
		
		if (isCleanRequired) {
			cleanInstances();
		}
		
		return instances;
	}
	
	
	
	private static void cleanInstances() {
		List<WeakReference<SessionManager>> refsToDelete = new ArrayList<WeakReference<SessionManager>>();
		
		for (WeakReference<SessionManager> sessionManagerRef : INSTANCE_LIST) {
			SessionManager sessionManager = sessionManagerRef.get();
			if (sessionManager == null) {
				refsToDelete.add(sessionManagerRef);
			}
		}
		
		for (WeakReference<SessionManager> sessionManagerRef : refsToDelete) {
			INSTANCE_LIST.remove(sessionManagerRef);
		}
	}
	
	
	
	
	/**
	 * retrieves a session
	 * 
	 * @param httpClient   the assigned http client
	 * @param domain       the domain
	 * @param path         the path
	 * @param create       true, if a new one should be created if not exits 
	 * @return the session
	 */
	IHttpSession getSession(HttpClient httpClient, String domain, String path, boolean create) {
    	
		for (DomainPathPair pair : sessions.keySet()) {
	    
	    	if (domain.equals(pair.getDomain()) || domain.endsWith(pair.getDomain())) {
	    		
	    		HttpSession session = sessions.get(pair);
	    		if (session != null) {
	    			session.setLastAccessTime(System.currentTimeMillis());
	    		}
	    		return session;
	    	}
		}	    	
    	
		if (create) {
			DomainPathPair pair = new DomainPathPair(domain, path);
			HttpSession session = new HttpSession(UUID.randomUUID().toString(), httpClient);
			sessions.put(pair, session);
			
			return session;
			
		} else {
			return null;
		}
    }
	
	
	
	
	

    
    @SuppressWarnings("unchecked")
	void clean() {
    	
    	try {
	    	HashMap<String, HttpSession> sessionsCopy = null;
	    	synchronized (sessions) {
	    		sessionsCopy = (HashMap<String, HttpSession>) sessions.clone(); 
			}
	    	
	    	long currentMillis = System.currentTimeMillis();
	    	for (Entry<String, HttpSession> entry : sessionsCopy.entrySet()) {
				if (entry.getValue().checkMaxIntervalExceeded(currentMillis)) {
		    		if (LOG.isLoggable(Level.FINE)) {
		    			LOG.fine("session " + entry.getValue() + " has been expired. Deleting it");
		    		}
					synchronized (sessions) {
			    		sessions.remove(entry.getKey()); 
					}
				}
			}
    	} catch (Exception e) {
    	    // eat and log exception
    		if (LOG.isLoggable(Level.FINE)) {
    			LOG.fine("error occured by cleaning sessions " + e.toString());
    		}
    	}
    }
	
	
	
	

	private static final class HttpSession implements IHttpSession {

	private static final long serialVersionUID = -5876183451306149950L;

		private final Map<String, Object> attributes = new HashMap<String, Object>();

		private boolean isValid = true;


		private String id = null;
		private long creationTime = System.currentTimeMillis();
		private long lastAccesTime = creationTime;
		private int maxInactiveIntervalSec = -1;
		
		private WeakReference<HttpClient> httpClientRef = null;
		
		
		public HttpSession(String id, HttpClient httpClient) {
			this.id = id;
			httpClientRef = new WeakReference<HttpClient>(httpClient);
		}
		
		public String getId() {
			return id;
		}
		
		
		public long getCreationTime() {
			return creationTime;
		}
		
		public long getLastAccessedTime() {
			if (!isValid()) {
				throw new IllegalStateException("Session " + getId() + " is invalid"); 
			}

			return lastAccesTime;
		}
		
		void setLastAccessTime(long lastAccesTime) {
			this.lastAccesTime = lastAccesTime;
		}
		

	    public Object getAttribute(String name) {
			if (!isValid()) {
				throw new IllegalStateException("Session " + getId() + " is invalid"); 
			}

	    	return attributes.get(name);
	    }
	    
	    

	    public void setAttribute(String name, Object value) {
			if (!isValid()) {
				throw new IllegalStateException("Session " + getId() + " is invalid"); 
			}

	    	attributes.put(name, value);
	    }
	    
	    

	    public void removeAttribute(String name) {
			if (!isValid()) {
				throw new IllegalStateException("Session " + getId() + " is invalid"); 
			}

	    	attributes.remove(name);
	    }

	    

	    @SuppressWarnings("unchecked")
		public Enumeration getAttributeNames() {
			if (!isValid()) {
				throw new IllegalStateException("Session " + getId() + " is invalid"); 
			}

	    	return Collections.enumeration(getAttributeNameSet());
	    }
	    

	    public Set<String> getAttributeNameSet() {
			if (!isValid()) {
				throw new IllegalStateException("Session " + getId() + " is invalid"); 
			}

	    	return Collections.unmodifiableSet(attributes.keySet());
	    }
	    
	    
	    public void setMaxInactiveInterval(int maxInactiveIntervalSec) {
	    	this.maxInactiveIntervalSec = maxInactiveIntervalSec;
	    }
	    
	    
	    public int getMaxInactiveInterval() {
	    	return maxInactiveIntervalSec;
	    }
	    
	    
	    public void invalidate() {
	    	isValid = false;
	    	httpClientRef = null;
	    }
	    
	    public boolean isValid() {
	    	return isValid;
	    }
	    
	    
	    boolean checkMaxIntervalExceeded(long currentMillis) {
	    	if (httpClientRef == null) {
	    		invalidate();
	    		return true;
	    	}
	    	
	    	HttpClient httpClient = httpClientRef.get(); 
	    	if (httpClient == null) {
	    		invalidate();
	    		return true;
	    		
	    	} else {
	    		if (System.currentTimeMillis() > (httpClient.getLastTimeRequestSentMillis() + (((long ) maxInactiveIntervalSec) * 1000))) {
	    			invalidate();
	    			return true;
	    		}
	    	}
	    	return false;
	    }
	    


	    
	    
	    @Override
	    public String toString() {
	    	StringBuilder sb = new StringBuilder();
	    	
	    	for (Entry<String, Object> entry : attributes.entrySet()) {
				sb.append(entry.getKey() + "=" + entry.getValue() + "&");
			}
	    	
	    	return sb.toString();
	    }
	}
	
	
	private static final class DomainPathPair {

		
		private String domain = "";
		private String path = "";
		
		
		DomainPathPair(String domain, String path) {
			this.domain = domain;
			
			if (path != null) {
				this.path = path;
			}
		}
		
		
		String getDomain() {
			return domain;
		}
		
		String getPath() {
			return path;
		}
		
		
		
		@Override
		public int hashCode() {
			return (domain + path).hashCode();
		}
		
		
		@Override
		public boolean equals(Object obj) {
			if (obj instanceof DomainPathPair) {
				DomainPathPair other = (DomainPathPair) obj;
				if (other.domain.equals(this.domain) && other.path.equals(this.path)) {
					return true;
				}
			}
			
			return false;
		}
	}

}
